# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import TestDefinitionClass
from .....schema_classes import TestDefinitionTypeClass
from .....schema_classes import TestInfoClass
from .....schema_classes import TestResultClass
from .....schema_classes import TestResultTypeClass
from .....schema_classes import TestResultsClass


TestDefinition = TestDefinitionClass
TestDefinitionType = TestDefinitionTypeClass
TestInfo = TestInfoClass
TestResult = TestResultClass
TestResultType = TestResultTypeClass
TestResults = TestResultsClass

# fmt: on
